#~ Attention: current version is prerelease!

#~ '(C) Links System Software 2009
#~ 'Feedback information
#~ 'www: http://lss2008.livejournal.com/
#~ 'E-mail1: designer@ls-software.ru
#~ 'E-mail2: kirill2007_77@mail.ru (search this e-mail to add skype contact)
#~ 'icq: 328-958-369

#~ ctrlpnts.rb prerelease ver. 1.6. 12-Jan-10
#~ Plug-in, which adds control construction points to curves and/or groups of faces

#~ How to use:
#~ 1)	put ctrlpnts.rb to Google SketchUp 6\Plugins folder
#~ 2)	start SketchUp 6

#~ Curve shape controlling
#~ 1)	draw arc curve, freehand curve or make a curve from a set of edges (use Rick Wilson's 'weld.rb' plugin)
#~ 2)	select curve, right-click on it, then choose Add control points to the curve command from the context menu
#~ 3) choose control points quantity and behaviour in the dialog box
#~ 4) then you are able to affect curve shape by moving added control points

#~ Geometry group shape controlling
#~ 1) make a group, which contains faces (ver. 1.2. supports only faces yet)
#~ 2) select the group, right-click on it, then choose "Add control points to group manually..." command from the context menu
#~ 3) put some construction control points somewhere on group or near it
#~    in ver. 1.5 or higher it is also possible to add control points around bounding box of the group automatically
#~    (select the group, right-click on it, then choose "Add control points to group bounds..." command from the context menu)
#~ 4) specify if you want to leave or not the initial geometry group and the value of the "Power" option
#~ 5) then you are able to affect geometry group shape by moving added control points


require 'sketchup.rb'

class Lss_C_Pnt_Interactive
  attr_accessor :c_pt_interactive_tool
  
  def initialize
    @entities=Sketchup.active_model.active_entities
    @under_cursor=nil
    @selected_c_pt=nil
    @ip = Sketchup::InputPoint.new
    @ip1 = Sketchup::InputPoint.new
  end
  
  def activate
    @under_cursor=nil
    @selected_c_pt=nil
    @ip = Sketchup::InputPoint.new
    @ip1 = Sketchup::InputPoint.new
    @stop_move=false
  end
  
  def reset
    @under_cursor=nil
    @selected_c_pt=nil
    @stop_move=false
    @face_mesh2draw=nil
  end
  
  def deactivate(view)
    self.reset
  end
  
  def get_c_pt_attrs
    @geom_groupID=@selected_c_pt.get_attribute "on_change_event_params","@geom_groupID"
    @deform_type=@selected_c_pt.get_attribute "on_change_event_params","@deform_type"
    @discard_mats=@selected_c_pt.get_attribute "on_change_event_params","@discard_mats"
    @pwr=@selected_c_pt.get_attribute "on_change_event_params","@pwr"
    @leave_init_grp=@selected_c_pt.get_attribute "on_change_event_params","@leave_init_grp"
  end
  
  def set_c_pt_attrs
    @selected_c_pt.set_attribute "on_change_event_params","@geom_groupID",@geom_group.entityID
    @selected_c_pt.set_attribute "on_change_event_params","@pwr",@pwr
    @selected_c_pt.set_attribute "on_change_event_params","@leave_init_grp",@leave_init_grp
    @selected_c_pt.set_attribute "on_change_event_params","old_x", @old_pos.x
    @selected_c_pt.set_attribute "on_change_event_params","old_y", @old_pos.y
    @selected_c_pt.set_attribute "on_change_event_params","old_z", @old_pos.z
  end
  def calc_mesh2draw
    @face_mesh2draw=Geom::PolygonMesh.new
    @face_mesh.points.each{|pt| @face_mesh2draw.add_point(pt)}
    p_gon_no=1
    while p_gon_no<=@face_mesh.polygons.length
      @face_mesh2draw.add_polygon(@face_mesh.polygon_points_at(p_gon_no))
      p_gon_no+=1
    end
    @offset_vec=@selected_c_pt.position.vector_to(@ip.position)
    if @offset_vec
      if @offset_vec.length>0
        offset_len=@offset_vec.length
        pt_ind=1
        while pt_ind<=@face_mesh.count_points
          pnt=@face_mesh.point_at(pt_ind)
          @grp_deform_inst.old_pos=@selected_c_pt.position
          @grp_deform_inst.hyp_r_pwr(pnt.transform(@geom_group.transformation), @pnts_arr)
          offset_coeff=@grp_deform_inst.offset_coeff
          curr_offset_vec=Geom::Vector3d.new(@offset_vec)
          curr_offset_vec.length=offset_len*offset_coeff
          if pnt
            pnt.offset!(curr_offset_vec) 
            @face_mesh2draw.set_point(pt_ind, pnt)
          end
          pt_ind+=1
        end
      end
    end
  end
  
  def onMouseMove(flags, x, y, view)
    @ip1.pick view, x, y
    if( @ip1 != @ip )
      view.invalidate
      @ip.copy! @ip1
      view.tooltip = @ip.tooltip
    end
  
    if @selected_c_pt.nil?
      ph=view.pick_helper
      ph.do_pick x,y
      @under_cursor = ph.best_picked
      if @under_cursor
        Sketchup.set_status_text("#{@under_cursor.typename} is under cursor.")
        view.invalidate
      else
        Sketchup.set_status_text("Click on control c_point to interact with.")
      end
    else
      if not @stop_move
        self.calc_mesh2draw
      else
        @selected_c_pt=nil
        @face_mesh2draw=nil
        view.invalidate
      end
    end
  end
  
  def onKeyDown(key, repeat, flags, view)
    if @under_cursor or @selected_c_pt
      if( key == VK_UP && repeat == 1 )
        @pwr+=1
        @selected_c_pt.set_attribute "on_change_event_params","@pwr",@pwr if @selected_c_pt
        @under_cursor.set_attribute "on_change_event_params","@pwr",@pwr if @under_cursor
        @grp_deform_inst=Lss_Ctrl_Pnts_Group.new(@geom_groupID,@pwr,@leave_init_grp,@deform_type)
        self.calc_mesh2draw if @selected_c_pt
        view.invalidate
      end
      if( key == VK_DOWN && repeat == 1 )
        @pwr+=-1
        @selected_c_pt.set_attribute "on_change_event_params","@pwr",@pwr if @selected_c_pt
        @under_cursor.set_attribute "on_change_event_params","@pwr",@pwr if @under_cursor
        @grp_deform_inst=Lss_Ctrl_Pnts_Group.new(@geom_groupID,@pwr,@leave_init_grp,@deform_type)
        self.calc_mesh2draw if @selected_c_pt
        view.invalidate
      end
    end
    
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
      @shift_down_time = Time.now
      
      # if we already have an inference lock, then unlock it
      if( view.inference_locked? )
          # calling lock_inference with no arguments actually unlocks
          view.lock_inference
          view.lock_inference @ip1
      end
    end
  end
  
  def draw_pwr_text(pt,view)
    txt_pt = view.screen_coords(pt)+[10,-15]
    view.draw_text(txt_pt, "power=#{@pwr}")
  end
  
  def draw(view)
    if @under_cursor
      if @under_cursor.typename=="ConstructionPoint"
        if @under_cursor.respond_to? "get_attribute"
          Sketchup.set_status_text("Fell free to click on this c_point to interact with.")
          pt=@under_cursor.position
          @geom_groupID=@under_cursor.get_attribute "on_change_event_params","@geom_groupID"
          if @geom_groupID
            @pwr=@under_cursor.get_attribute "on_change_event_params","@pwr"
            self.draw_pwr_text(pt,view)
            self.draw_c_pt(pt,view)
            @entities.each{|ent|
              if ent.respond_to? "get_attribute"
                @geom_group=ent if ent.entityID==@geom_groupID
              end
            }
            self.draw_yellow_grp_bnds(@geom_group.bounds,view)
          end
        end
      else
        bnds=@under_cursor.bounds
        self.draw_bnds(bnds,view) if bnds
      end
    end
    if @face_mesh2draw
      polygon_no=1
      while polygon_no<=@face_mesh2draw.polygons.length
        pnts=@face_mesh2draw.polygon_points_at(polygon_no)
        pnts.collect!{|pnt| pnt.transform!(@geom_group.transformation)} if @geom_group and not @geom_group.deleted?
        status = view.drawing_color="gray"
        status=view.draw_polyline pnts
        polygon_no+=1
      end
    end
    if @ip and @selected_c_pt
      self.draw_pwr_text(@ip.position,view)
      self.draw_vec(@selected_c_pt.position,@ip.position,"..",view)
      view.draw_points @ip.position,(@pwr*3).round,7,"green"
    end
  end

  def draw_bnds(bnds,view)
    corner_no=0
    while corner_no<=7
      view.draw_points bnds.corner(corner_no), 6, 4,"red"
      corner_no+=1
    end
  end 
  
  def draw_yellow_grp_bnds(bnds,view)
    corner_no=0
    while corner_no<=7
      view.draw_points bnds.corner(corner_no), 6, 2,"yellow"
      corner_no+=1
    end
  end
  
  def draw_c_pt(pt,view)
    view.draw_points pt, (@pwr*3).round, 2,"green"
  end
  
  def make_polygon_mesh
    @face_mesh=Geom::PolygonMesh.new
    @geom_group.entities.each{|fc|
      mesh_pnts=Array.new
      if fc.typename=="Face"
        single_face_mesh=fc.mesh
        single_face_mesh.points.each{|pnt|
          mesh_pnts+=[pnt]
        }
        mesh_pnts.uniq!
        mesh_pnts.each{|pnt| @face_mesh.add_point(pnt)}
        polygon_ind=1
        while polygon_ind<=single_face_mesh.polygons.length
          face_pnts=single_face_mesh.polygon_points_at(polygon_ind)
          begin
            @face_mesh.add_polygon(face_pnts)
          rescue
            puts "Failed to add polygon"
          end
          polygon_ind+=1
        end
      end
    }
  end
  
  def onLButtonUp(flags, x, y, view)
    @stop_move=false
    if @selected_c_pt.nil?
      @c_pnts_arr=Array.new
      @pnts_arr=Array.new
      ph=view.pick_helper
      ph.do_pick x,y
      @selected_c_pt=ph.best_picked
      if @selected_c_pt
        if @selected_c_pt.typename=="ConstructionPoint"
          @under_cursor=nil
          if @selected_c_pt.respond_to? "get_attribute"
            @geom_groupID=@under_cursor.get_attribute "on_change_event_params","@geom_groupID" if @under_cursor
            if @geom_groupID
              @entities.each{|ent|
                if ent.respond_to? "get_attribute"
                  chk_grpID=ent.get_attribute "on_change_event_params","@geom_groupID"
                  if @geom_groupID==chk_grpID
                    @c_pnts_arr+=[ent] 
                    old_pos_x=ent.get_attribute "on_change_event_params","old_x"
                    old_pos_y=ent.get_attribute "on_change_event_params","old_y"
                    old_pos_z=ent.get_attribute "on_change_event_params","old_z"
                    old_pos=Geom::Point3d.new(old_pos_x,old_pos_y,old_pos_z) if old_pos_x and old_pos_y and old_pos_z
                    @pnts_arr+=[old_pos]
                  end
                  @geom_group=ent if ent.entityID==@geom_groupID
                  self.make_polygon_mesh
                  self.get_c_pt_attrs
                  @grp_deform_inst=Lss_Ctrl_Pnts_Group.new(@geom_groupID,@pwr,@leave_init_grp,@deform_type)
                end
              }
            end
          end
        else
          @selected_c_pt=nil
        end
      end
    else
      c_pnt=@entities.add_cpoint @ip.position
      self.get_c_pt_attrs
      c_pnt.set_attribute "on_change_event_params","@geom_groupID",@geom_group.entityID
      c_pnt.set_attribute "on_change_event_params","@pwr", @pwr
      c_pnt.set_attribute "on_change_event_params","@leave_init_grp", @leave_init_grp
      c_pnt.set_attribute "on_change_event_params","@discard_mats", @discard_mats
      c_pnt.set_attribute "on_change_event_params","@deform_type", @deform_type
      c_pnt_observer=Lss_C_Pnts_Group_EntityObserver.new(c_pnt)
      c_pnt.add_observer(c_pnt_observer)
      c_pnt.set_attribute "on_change_event_params","old_x", @selected_c_pt.position.x
      c_pnt.set_attribute "on_change_event_params","old_y", @selected_c_pt.position.y
      c_pnt.set_attribute "on_change_event_params","old_z", @selected_c_pt.position.z
      @selected_c_pt.erase!
      c_pnt_observer.onChangeEntity(c_pnt)
      self.reset
      Sketchup.active_model.select_tool @c_pt_interactive_tool if @discard_mats=="Yes"
    end
    view.invalidate
  end
  
  def onCancel(reason, view)
    if reason==0
      @stop_move=true
    end
  end
  
  def draw_vec(pt1,pt2,stpl,view)
    view.line_stipple = stpl
    view = view.set_color_from_line pt1,pt2
    status = view.draw_lines pt1, pt2
  end
  
end #class Lss_C_Pnt_Interactive

class Lss_C_Pnts_AppObserver < Sketchup::AppObserver #works good but slows down a bit huge models opening...
   def onOpenModel(model)
      @entities=Sketchup.active_model.entities
      @entities.each{|c_pnt|
        if c_pnt.typename=="ConstructionPoint"
          if c_pnt.respond_to? "get_attribute"
            chk_curve_edge0_ID=c_pnt.get_attribute "on_change_event_params","@curve_edge0_ID"
            if not chk_curve_edge0_ID.nil?
              ent_no=0
              timer_id=UI.start_timer(0.01,true) {
                edg=@entities[ent_no]
                if edg
                  if edg.typename=="Edge"
                    if edg.start.position==c_pnt.position or edg.end.position==c_pnt.position
                      edg_curve=edg.curve
                      if edg_curve
                        curve_edge0_ID=edg_curve.edges[0].entityID
                        c_pnt.set_attribute "on_change_event_params","@curve_edge0_ID",curve_edge0_ID
                        c_pnt.add_observer(Lss_C_Pnts_EntityObserver.new(c_pnt))
                      else
                        puts "Failed to link control c-point, because linked curve was exploded..."
                      end
                    end
                  end
                end
                timer_id=UI.stop_timer timer_id if ent_no>=@entities.length
                ent_no+=1
              }
            end
          end
        end
      }
   end
 end #class Lss_C_Pnts_AppObserver


class Lss_C_Pnts_Group_EntityObserver < Sketchup::EntityObserver 
    def initialize(ctrl_c_pnt)
      @entities=Sketchup.active_model.active_entities
      @geom_groupID=ctrl_c_pnt.get_attribute "on_change_event_params","@geom_groupID"
      @deform_type=ctrl_c_pnt.get_attribute "on_change_event_params","@deform_type"
      @initial_pos=ctrl_c_pnt.position
      @old_pos=ctrl_c_pnt.position
      @on_change_cnt=0
      @offset_vec=Geom::Vector3d.new
      ctrl_c_pnt.set_attribute "on_change_event_params","init_x", @initial_pos.x
      ctrl_c_pnt.set_attribute "on_change_event_params","init_y", @initial_pos.y
      ctrl_c_pnt.set_attribute "on_change_event_params","init_z", @initial_pos.z
      ctrl_c_pnt.set_attribute "on_change_event_params","old_x", @old_pos.x
      ctrl_c_pnt.set_attribute "on_change_event_params","old_y", @old_pos.y
      ctrl_c_pnt.set_attribute "on_change_event_params","old_z", @old_pos.z
    end
    def onChangeEntity(entity)
      @discard_mats=entity.get_attribute "on_change_event_params","@discard_mats"
      @geom_groupID=entity.get_attribute "on_change_event_params","@geom_groupID"
      @pwr=entity.get_attribute "on_change_event_params","@pwr"
      @leave_init_grp=entity.get_attribute "on_change_event_params","@leave_init_grp"
      @redraw_group=Lss_Ctrl_Pnts_Group.new(@geom_groupID,@pwr,@leave_init_grp,@deform_type)
      Sketchup.active_model.select_tool @redraw_group
      @on_change_cnt+=1
      @new_pos=entity.position
      @offset_vec=@old_pos.vector_to(@new_pos)
      if @discard_mats=="No"
        @redraw_group.deform_group(@old_pos, @new_pos, @offset_vec)
      else
        @redraw_group.deform_group_discard_mats(@old_pos, @new_pos, @offset_vec)
      end
      @old_pos=entity.position
      #~ entity.set_attribute "on_change_event_params","old_x", @old_pos.x
      #~ entity.set_attribute "on_change_event_params","old_y", @old_pos.y
      #~ entity.set_attribute "on_change_event_params","old_z", @old_pos.z
    end
end #class Lss_C_Pnts_Group_EntityObserver 

class Lss_Tiny_Progr_Bar
  attr_accessor :percents_ready
  attr_accessor :progr_string
  def initialize(tot_cnt,progr_char,rest_char,scale_coeff)
    @scale_coeff=scale_coeff
    @scale_coeff=2 if @scale_coeff==nil or @scale_coeff==0
    if tot_cnt
      @tot_cnt=tot_cnt
      @tot_cnt=1 if @tot_cnt==0
    end
    @progr_char=progr_char
    @rest_char=rest_char
    @percents_ready=0
  end
  
  def update(curr_cnt)
    @curr_cnt=curr_cnt
    @percents_ready=(100*@curr_cnt/(@tot_cnt)).round
    progr_str=@progr_char*((@percents_ready/@scale_coeff).round)+@rest_char*(100/@scale_coeff-(@percents_ready/@scale_coeff).round)
    @progr_string="#{@percents_ready}% #{progr_str}"
  end
  
end #class Lss_Tiny_Progr_Bar

class Lss_Ctrl_Pnts_Group
  attr_accessor :offset_coeff
  attr_accessor :old_pos
  
  def initialize (geom_grpID,pwr,leave_init_grp,deform_type)
    @deform_type=deform_type
    @leave_init_grp=leave_init_grp
    @geom_groupID=geom_grpID
    @pwr=pwr
    @entities=Sketchup.active_model.active_entities ####### Tig, thanx
    @offset_coeff=1
  end
  
 def activate
    @break_timer=false
  end
  
 def draw(view)
   if @geom_group and @face_ind
      if @face_ind and @geom_group.entities.length>0
        tot_cnt=@geom_group.entities.length
        percents_ready=(100*@face_ind/(tot_cnt)).round
        txt_pt3d=Geom::Point3d.new(@geom_group.bounds.max)
        txt_pt=view.screen_coords(txt_pt3d)
        progr_char="|"
        rest_char="_"
        progr_string=progr_char*((percents_ready/5).round)+rest_char*(20-(percents_ready/5).round)
        txt_string="#{percents_ready}% #{progr_string}"
        view.draw_text(txt_pt, txt_string)
      end
    end
  end
  
  def onCancel(reason, view)
    if reason==0
      @break_timer=true
    end
  end
  
  def hyp_r_pwr(pnt, pt_arr)
    dist=@old_pos.distance pnt
    if dist>0
      hyp_r_pwr_sum=0
      pt_arr.each{|pt|
        dist1=pnt.distance pt
        if dist1>0 
          hyp_r_pwr_sum+=((1.0)/(dist1))**@pwr if not pt==@new_pos
        else
          @offset_coeff=0
          return
        end
      }
      hyp_r_pwr=((1.0)/(dist))**@pwr
      @offset_coeff=1-(hyp_r_pwr_sum-hyp_r_pwr)/(hyp_r_pwr_sum)
    else
      @offset_coeff=1
    end
  end
  
  def deform_group_discard_mats(old_pos, new_pos, offset_vec)
    model = Sketchup.active_model 
    status = model.start_operation "Geometry group deformation"
    @geom_group=nil
    @offset_len=offset_vec.length if offset_vec
    @c_pnts_arr=Array.new
    @pnts_arr=Array.new
    @entities.each{|ent|
      if ent.respond_to? "get_attribute"
        chk_grpID=ent.get_attribute "on_change_event_params","@geom_groupID"
        if @geom_groupID==chk_grpID
          @c_pnts_arr+=[ent] 
          old_pos_x=ent.get_attribute "on_change_event_params","old_x"
          old_pos_y=ent.get_attribute "on_change_event_params","old_y"
          old_pos_z=ent.get_attribute "on_change_event_params","old_z"
          old_pos=Geom::Point3d.new(old_pos_x,old_pos_y,old_pos_z) if old_pos_x and old_pos_y and old_pos_z
          @pnts_arr+=[old_pos]
        end
        @geom_group=ent if ent.entityID==@geom_groupID
      end
    }
    if @geom_group.nil?
      UI.messagebox "Most probably geometry group was erased"
      return
    end

    @new_group=Sketchup.active_model.active_entities.add_group ############# Tig, thanx
    @new_group.transform!(@geom_group.transformation)
    face_mesh=Geom::PolygonMesh.new
    @geom_group.entities.each{|fc|
      mesh_pnts=Array.new
      if fc.typename=="Face"
        single_face_mesh=fc.mesh
        single_face_mesh.points.each{|pnt|
          mesh_pnts+=[pnt]
        }
        mesh_pnts.uniq!
        mesh_pnts.each{|pnt| face_mesh.add_point(pnt)}
        polygon_ind=1
        while polygon_ind<=single_face_mesh.polygons.length
          face_pnts=single_face_mesh.polygon_points_at(polygon_ind)
          begin
            face_mesh.add_polygon(face_pnts)
          rescue
            puts "Failed to add polygon"
          end
          polygon_ind+=1
        end
      end
    }
    
    tot_cnt=face_mesh.points.length
    if tot_cnt
      prgr_bar=Lss_Tiny_Progr_Bar.new(tot_cnt,"|","_",2)
    end
    curr_c_pt_no=0
    tot_c_pt_cnt=@c_pnts_arr.length
    
    @c_pnts_arr.each{|c_pnt|
      old_pos_x=c_pnt.get_attribute "on_change_event_params","old_x"
      old_pos_y=c_pnt.get_attribute "on_change_event_params","old_y"
      old_pos_z=c_pnt.get_attribute "on_change_event_params","old_z"
      old_pos=Geom::Point3d.new(old_pos_x,old_pos_y,old_pos_z)
      @old_pos=old_pos
      @new_pos=c_pnt.position
      @offset_vec=@old_pos.vector_to(@new_pos)
      offset_len=@offset_vec.length
      if offset_len>0
        pt_ind=1
        while pt_ind<=face_mesh.points.length
          mesh_pt=face_mesh.point_at(pt_ind)
          self.hyp_r_pwr(mesh_pt.transform(@geom_group.transformation),@pnts_arr)
          curr_offset_vec=Geom::Vector3d.new(@offset_vec)
          curr_offset_vec.length=offset_len*@offset_coeff
          mesh_pt.offset!(curr_offset_vec)
          face_mesh.set_point(pt_ind,mesh_pt)
          prgr_bar.update(pt_ind)
          Sketchup.set_status_text "Processing c_point/total count: #{curr_c_pt_no}/#{tot_c_pt_cnt} #{prgr_bar.progr_string}"
          pt_ind+=1
        end
        c_pnt.set_attribute "on_change_event_params","old_x", @new_pos.x
        c_pnt.set_attribute "on_change_event_params","old_y", @new_pos.y
        c_pnt.set_attribute "on_change_event_params","old_z", @new_pos.z
        old_pos_ind=@pnts_arr.index(@old_pos)
        @pnts_arr[old_pos_ind]=@new_pos
      end
      curr_c_pt_no+=1
    }
    Sketchup.set_status_text "C_points processing finished."
    
    @new_group.entities.add_faces_from_mesh(face_mesh,12)
    
    @c_pnts_arr.each{|c_pnt| 
      c_pnt.set_attribute "on_change_event_params","@geom_groupID",@new_group.entityID
    }
    @geom_group.erase! if @leave_init_grp=="No"
    Sketchup.set_status_text "Deformation finished."
    status = model.commit_operation
  end
  
  def deform_group (old_pos, new_pos, offset_vec)
    model = Sketchup.active_model 
    status = model.start_operation "Geometry group deformation"
    view=model.active_view
    @geom_group=nil
    @offset_len=offset_vec.length if offset_vec
    @c_pnts_arr=Array.new
    @pnts_arr=Array.new
    @entities.each{|ent|
      if ent.respond_to? "get_attribute"
        chk_grpID=ent.get_attribute "on_change_event_params","@geom_groupID"
        if @geom_groupID==chk_grpID
          @c_pnts_arr+=[ent] 
          old_pos_x=ent.get_attribute "on_change_event_params","old_x"
          old_pos_y=ent.get_attribute "on_change_event_params","old_y"
          old_pos_z=ent.get_attribute "on_change_event_params","old_z"
          old_pos=Geom::Point3d.new(old_pos_x,old_pos_y,old_pos_z) if old_pos_x and old_pos_y and old_pos_z
          @pnts_arr+=[old_pos]
        end
        @geom_group=ent if ent.entityID==@geom_groupID
      end
    }
    if @geom_group.nil?
      UI.messagebox "Most probably geometry group was erased"
      return
    end
    @new_group=Sketchup.active_model.active_entities.add_group ############ Tig, thanx
    @new_group.transform!(@geom_group.transformation)
    @face_ind=0
    timer_id=UI.start_timer(0.01,true) {
      view.invalidate
      ent=@geom_group.entities[@face_ind]
      if not ent.nil?
        if ent.typename=="Face"
          pnts_arr_temp=Array.new
          @pnts_arr.each{|pnt| pnts_arr_temp+=[Geom::Point3d.new(pnt)]}
          fc_mesh=ent.mesh
          pt_ind=1
          while pt_ind<=fc_mesh.count_points
            pnt=fc_mesh.point_at(pt_ind)
            @c_pnts_arr.each{|c_pt|
              old_pos_x=c_pt.get_attribute "on_change_event_params","old_x"
              old_pos_y=c_pt.get_attribute "on_change_event_params","old_y"
              old_pos_z=c_pt.get_attribute "on_change_event_params","old_z"
              old_pos=Geom::Point3d.new(old_pos_x,old_pos_y,old_pos_z)
              @old_pos=old_pos
              @new_pos=c_pt.position
              @offset_vec=@old_pos.vector_to(@new_pos)
              offset_len=@offset_vec.length
              if offset_len>0
                self.hyp_r_pwr(pnt.transform(@geom_group.transformation),@pnts_arr)
                curr_offset_vec=Geom::Vector3d.new(@offset_vec)
                curr_offset_vec.length=offset_len*@offset_coeff
                pnt.offset!(curr_offset_vec)
                fc_mesh.set_point(pt_ind, pnt)
                old_pos_ind=@pnts_arr.index(@old_pos)
                @pnts_arr[old_pos_ind]=@new_pos
              end
            }
            pt_ind+=1
            @pnts_arr=Array.new
            pnts_arr_temp.each{|pnt| @pnts_arr+=[Geom::Point3d.new(pnt)]}
          end
          if @new_group and not @new_group.deleted?
            @new_group.entities.add_faces_from_mesh(fc_mesh,12,ent.material,ent.back_material)
          else
            puts "Operation aborted."
            status = model.commit_operation
            timer_id=UI.stop_timer timer_id 
          end
        end
      end
      ent.visible=false if @leave_init_grp=="No"
      @face_ind+=1
      if @face_ind>=@geom_group.entities.length
        timer_id=UI.stop_timer timer_id 
        @c_pnts_arr.each{|c_pnt| 
          c_pnt.set_attribute "on_change_event_params","@geom_groupID",@new_group.entityID
          c_pnt.set_attribute "on_change_event_params","old_x", c_pnt.position.x
          c_pnt.set_attribute "on_change_event_params","old_y", c_pnt.position.y
          c_pnt.set_attribute "on_change_event_params","old_z", c_pnt.position.z
        }
        @geom_group.erase! if @leave_init_grp=="No"
        status = model.commit_operation
      end
      if @break_timer
        @geom_group.entities.each{|ent| ent.visible=true}
        timer_id=UI.stop_timer timer_id
        @new_group.erase!
        Sketchup.active_model.select_tool nil
        status = model.commit_operation
        puts "Group deformation was cancelled."
      end
    }
  end
  
  def deactivate(view)
    @face_ind=nil
    view.invalidate
  end
end #class Lss_Ctrl_Pnts_Group

class Lss_Ctrl_Pnts_Sel_Tool
    def initialize
      @entities=Sketchup.active_model.active_entities
      @selection=Sketchup.active_model.selection
      @geom_group=nil
      @b_p=nil
      @c_points_set=Set.new
      @selection.each{|c_pnt|
        if c_pnt.typename=="ConstructionPoint"
          @c_points_set.insert c_pnt
        end
      }
      @selection.clear
    end
    
    def onMouseMove(flags, x, y, view)
      ph=view.pick_helper
      ph.do_pick x,y
      @b_p = ph.best_picked
      if @b_p
        Sketchup.set_status_text("#{@b_p.typename} is under cursor.")
        view.invalidate
      else
        Sketchup.set_status_text("Click on group, which contains faces.")
      end
    end
    
    def draw(view)
      if @b_p
        if @b_p.typename=="Group"
          self.draw_grp_bnds(@b_p.bounds,view)
          self.highlight_c_pnts(view) if @c_points_set
        else
          txt_pt3d=Geom::Point3d.new(@b_p.bounds.center)
          txt_pt = view.screen_coords(txt_pt3d)
          view.draw_text(txt_pt, "This is '#{@b_p.typename}'")
          self.draw_no_grp_bnds(@b_p.bounds,view)
        end
      end
    end
    
    def draw_grp_bnds(grp_bounds,view)
      corner_no=0
      while corner_no<=7
        view.draw_points grp_bounds.corner(corner_no), 6, 1,"green"
        corner_no+=1
      end
    end
    
    def draw_no_grp_bnds(bnds,view)
      corner_no=0
      while corner_no<=7
        view.draw_points bnds.corner(corner_no), 6, 4,"red"
        corner_no+=1
      end
    end 
    
    def highlight_c_pnts(view)
      @c_points_set.each{|c_pt|
        view.draw_points c_pt.position, 4, 2,"yellow"
      }
    end
    
    def onLButtonUp(flags, x, y, view)
      ph=view.pick_helper
      ph.do_pick x,y
      @b_p = ph.best_picked
      if not @b_p.nil?
        puts "#{@b_p.typename} is picked."
        Sketchup.set_status_text("#{@b_p.typename} is picked. Click on group.")
        if @b_p.typename == "Group"
          @geom_group=@b_p
          Sketchup.active_model.select_tool nil 
          Sketchup.set_status_text("#{@b_p.typename} is picked. Choose controlling options now.")
        end
      else
        puts "Nothing is picked"
        Sketchup.set_status_text("Nothing is picked. Click on group.")
      end
    end
    
    def deactivate(view)
      view.invalidate
      prompts = ["Power: ", "Leave initial geometry group: ", "Discard materials: "]
      defaults= [3.0,"No", "No"]
      list = ["", "Yes|No", "Yes|No"]
      @pwr, @leave_init_grp, @discard_mats, @deform_type = UI.inputbox prompts, defaults, list, "Geometry Group Control Points Settings"
      @c_points_set.each{|c_pnt|
        c_pnt.set_attribute "on_change_event_params","@geom_groupID",@geom_group.entityID
        c_pnt.set_attribute "on_change_event_params","@pwr", @pwr
        c_pnt.set_attribute "on_change_event_params","@leave_init_grp", @leave_init_grp
        c_pnt.set_attribute "on_change_event_params","@discard_mats", @discard_mats
        c_pnt.set_attribute "on_change_event_params","@deform_type", @deform_type
        c_pnt.add_observer(Lss_C_Pnts_Group_EntityObserver.new(c_pnt))
      }
    end
end #class Lss_Ctrl_Pnts_Sel_Tool

class Lss_Ctrl_Pnts_Add_Tool
    def initialize
      @entities=Sketchup.active_model.active_entities ########## Tig, thanx
      @selection=Sketchup.active_model.selection
      @geom_group=@selection[0]
      @pwr=3.0
      @leave_init_grp="No"
      @ip=nil
      @ip1=nil
    end
    
    def activate
      @ip = Sketchup::InputPoint.new
      @ip1 = Sketchup::InputPoint.new
    end
    
    def onMouseMove(flags, x, y, view)
      @c_pnt=nil
      @ip1.pick view, x, y
      if( @ip1 != @ip )
          view.invalidate
          @ip.copy! @ip1
          view.tooltip = @ip.tooltip
      end
    end
    
    def draw(view)
      if @ip.valid?
        if @ip.display?
          @ip.draw(view)
          self.draw_geom(@ip.position,view)
        end
      end
    end
    
    def draw_geom(p,view)
      view.draw_points p, 3, 3,"red"
    end
    
    def reset(view)
      @ip.clear
     if( view )
        view.tooltip = nil
        view.invalidate if @drawn
      end
    end
    
    def onLButtonUp(flags, x, y, view)
      @ip.pick view, x, y
      control_c_pnt=view.model.active_entities.add_cpoint @ip.position
      control_c_pnt.set_attribute "on_change_event_params","@geom_groupID",@geom_group.entityID
      control_c_pnt.set_attribute "on_change_event_params","@pwr",@pwr
      control_c_pnt.set_attribute "on_change_event_params","@leave_init_grp",@leave_init_grp
      #~ control_c_pnt.set_attribute "on_change_event_params","@discard_mats",@discard_mats
      control_c_pnt.add_observer(Lss_C_Pnts_Group_EntityObserver.new(control_c_pnt))
    end

    def onCancel(reason, view)
     puts "Add control points process was cancelled for reason #" + reason.to_s +
     " in view:" + view.to_s
     self.reset(view)
     Sketchup.active_model.select_tool nil
   end

    def deactivate(view)
      view.invalidate
      prompts = ["Power: ", "Leave initial geometry group: ", "Discard materials: "]
      defaults= [3.0,"No", "No"]
      list = ["", "Yes|No", "Yes|No"]
      @pwr, @leave_init_grp, @discard_mats, @deform_type = UI.inputbox prompts, defaults, list, "Geometry Group Control Points Settings"
      @entities.each{|ent|
        if ent.respond_to? "get_attribute"
          chk_grpID=ent.get_attribute "on_change_event_params","@geom_groupID"
          if @geom_group.entityID==chk_grpID
            ent.set_attribute "on_change_event_params","@pwr", @pwr
            ent.set_attribute "on_change_event_params","@leave_init_grp", @leave_init_grp
            ent.set_attribute "on_change_event_params","@discard_mats", @discard_mats
            ent.set_attribute "on_change_event_params","@deform_type", @deform_type
          end
        end
      }
    end
end #class Lss_Ctrl_Pnts_Add_Tool

class Lss_C_Pnts_EntityObserver < Sketchup::EntityObserver 
    def initialize(ctrl_c_pnt)
      @entities=Sketchup.active_model.active_entities
      @curve_edge0_ID=ctrl_c_pnt.get_attribute "on_change_event_params","@curve_edge0_ID"
      @entities.each{|ent|
        @curve=ent.curve if ent.entityID==@curve_edge0_ID
      }
      @vrt_ind=ctrl_c_pnt.get_attribute "on_change_event_params","vrt_ind"
      @skip_vert_cnt=ctrl_c_pnt.get_attribute "on_change_event_params","@skip_vert_cnt"
      @pnt_follow_type=ctrl_c_pnt.get_attribute "on_change_event_params","@pnt_follow_type"
      @redraw_curve=Lss_Ctrl_Pnts.new(@skip_vert_cnt,@pnt_follow_type,@curve)
      @on_change_cnt=0
      @init_crv_verts=Array.new
      @curve.vertices.each{|vrt| @init_crv_verts+=[vrt.position]}
    end
   def onChangeEntity(entity)
      @on_change_cnt+=1
      puts "Curve vertex change No: #{@on_change_cnt}"
      puts "Curve vertex index: #{@vrt_ind}"
     begin
        if not @curve.nil?
          @redraw_curve.redraw_crv(@vrt_ind,entity.position,@init_crv_verts)
        end
      rescue
        puts "onChangeEntity handle aborted"
        Sketchup.active_model.abort_operation
      end
      if not @pnt_follow_type=="iterative smoothed"
        begin
          if not @curve.nil?
            @entities.each{|c_pnt|
              if c_pnt.typename=="ConstructionPoint"
                if c_pnt.respond_to? "get_attribute"
                  chk_curve_edge0_ID=c_pnt.get_attribute "on_change_event_params","@curve_edge0_ID"
                  if chk_curve_edge0_ID==@curve_edge0_ID
                    new_vrt_ind=c_pnt.get_attribute "on_change_event_params","vrt_ind"
                    @redraw_curve.redraw_crv(new_vrt_ind,c_pnt.position,@init_crv_verts) 
                  end
                end
              end
            }
          end
        rescue
          puts "onChangeEntity handle aborted"
          Sketchup.active_model.abort_operation
        end
      end
    end 
  end #class Lss_C_Pnts_EntityObserver

class Lss_Ctrl_Pnts
  def initialize(skip_vert_cnt,pnt_follow_type,curve)
    @skip_vert_cnt=skip_vert_cnt
    @pnt_follow_type=pnt_follow_type
    @curve=curve
    if not @curve.vertices[0].position==@curve.vertices[-1].position
      @curve_is_closed=false
    else
      @curve_is_closed=true
    end
  end
  
  def add_c_pnts
    model = Sketchup.active_model 
    entities = model.active_entities 
    if @skip_vert_cnt==-1
      @skip_vert_cnt=@curve.vertices.length-2
    end
    c_pnt_ind=0
    vrt_ind=0
    if not @curve_is_closed
      curve_verts_cnt=@curve.vertices.length
    else
      curve_verts_cnt=@curve.vertices.length-1
    end
    while vrt_ind<curve_verts_cnt
      control_c_pnt=entities.add_cpoint @curve.vertices[vrt_ind].position
      control_c_pnt.set_attribute "on_change_event_params","@curve_edge0_ID",@curve.edges[0].entityID
      control_c_pnt.set_attribute "on_change_event_params","vrt_ind",vrt_ind
      control_c_pnt.set_attribute "on_change_event_params","@skip_vert_cnt",@skip_vert_cnt
      control_c_pnt.set_attribute "on_change_event_params","@pnt_follow_type",@pnt_follow_type
      control_c_pnt.add_observer(Lss_C_Pnts_EntityObserver.new(control_c_pnt))
      vrt_ind+=1+@skip_vert_cnt
      c_pnt_ind+=1
    end
  end
  
  def cosine_smoothed
    vrt_ind=0
    new_crv_verts=Array.new
    @curve.vertices.each{|vrt| new_crv_verts+=[vrt.position]}
    while vrt_ind<@curve.vertices.length 
      if (vrt_ind-@changed_vrt_ind).abs<=@skip_vert_cnt or ((@curve.vertices.length-1-vrt_ind).abs<=@skip_vert_cnt and @changed_vrt_ind==0 and @curve_is_closed)
        offset_vec=@curve.vertices[@changed_vrt_ind].position.vector_to @new_position
        ind_dist=(vrt_ind-@changed_vrt_ind).abs
        ind_dist=(@curve.vertices.length-1-vrt_ind).abs if @changed_vrt_ind==0 and @curve_is_closed and vrt_ind>@skip_vert_cnt
        vec_len=offset_vec.length
        mock_ang_step=(Math::PI)/(@skip_vert_cnt+1.0)
        mock_ang=ind_dist*mock_ang_step
        offset_coeff=(1+Math.cos(mock_ang))/2
        offset_vec.length=vec_len*offset_coeff if vec_len>0
        new_crv_verts[vrt_ind].offset!  offset_vec
      end
      vrt_ind+=1      
    end
    if @curve_is_closed
      new_crv_verts[-1]=new_crv_verts[0]
    end
    @curve.move_vertices new_crv_verts
  end
  
  def linear
    vrt_ind=0
    new_crv_verts=Array.new
    @curve.vertices.each{|vrt| new_crv_verts+=[vrt.position]}
    while vrt_ind<@curve.vertices.length 
      if (vrt_ind-@changed_vrt_ind).abs<=@skip_vert_cnt or ((@curve.vertices.length-1-vrt_ind).abs<=@skip_vert_cnt and @changed_vrt_ind==0 and @curve_is_closed)
        offset_vec=@curve.vertices[@changed_vrt_ind].position.vector_to @new_position
        ind_dist=(vrt_ind-@changed_vrt_ind).abs
        ind_dist=(@curve.vertices.length-1-vrt_ind).abs if @changed_vrt_ind==0 and @curve_is_closed and vrt_ind>@skip_vert_cnt
        offset_coeff=(@skip_vert_cnt+1.0-ind_dist)/(@skip_vert_cnt+1)
        vec_len=offset_vec.length
        offset_vec.length=vec_len*offset_coeff if vec_len>0
        new_crv_verts[vrt_ind].offset!  offset_vec
      end
      vrt_ind+=1      
    end
    if @curve_is_closed
      new_crv_verts[-1]=new_crv_verts[0]
    end
    @curve.move_vertices new_crv_verts
  end
  
  def total_length(pnts_ary)
    @tot_len=0
    vrt_ind=0
    while vrt_ind<pnts_ary.length-1
      vert1=pnts_ary[vrt_ind]; vert2=pnts_ary[vrt_ind+1]
      @tot_len+=vert1.distance vert2
      vrt_ind+=1
    end
  end
  
  def average_len(pnts_ary)
    total_length(pnts_ary)
    @aver_len=@tot_len/(pnts_ary.length)
  end
  
  def redraw_crv(changed_vrt_ind,new_position,init_crv_verts)
    @changed_vrt_ind=changed_vrt_ind
    @new_position=new_position
    @init_crv_verts=init_crv_verts
    @new_crv_verts=Array.new(init_crv_verts)
    model = Sketchup.active_model 
    entities = model.active_entities
    case @pnt_follow_type
      when "linear"
        linear
      when "cosine smoothed"
        cosine_smoothed
      #~ when "iterative smoothed"
        #~ iterative_smoothed
    end
  end
end #class Lss_Ctrl_Pnts

class Lss_Add_C_Pnts_BB
  def initialize
    @entities=Sketchup.active_model.active_entities ########## Tig, thanx
    @selection=Sketchup.active_model.selection
    @geom_group=@selection[0]
    @pwr=3.0
    @leave_init_grp="No"
  end
  
  def set_attrs
    model=Sketchup.active_model
    prompts = ["Power: ", "Leave initial geometry group: ", "Discard materials: "]
    defaults= [3.0,"No", "No"]
    list = ["", "Yes|No", "Yes|No"]
    @pwr, @leave_init_grp, @discard_mats, @deform_type = UI.inputbox prompts, defaults, list, "Geometry Group Control Points Settings"
    
    @entities.each{|ent|
      if ent.respond_to? "get_attribute"
        chk_grpID=ent.get_attribute "on_change_event_params","@geom_groupID"
        if @geom_group.entityID==chk_grpID
          ent.set_attribute "on_change_event_params","@pwr", @pwr
          ent.set_attribute "on_change_event_params","@leave_init_grp", @leave_init_grp
          ent.set_attribute "on_change_event_params","@discard_mats", @discard_mats
          ent.set_attribute "on_change_event_params","@deform_type", @deform_type
        end
      end
    }
    #~ status = model.commit_operation
  end
  
  def activate
    model=Sketchup.active_model
    #~ status=model.start_operation "B_box control points placing"
    prompts=["Place control points to: "]
    defaults=["bounding box corners"]
    list=["bounding box corners|centers of b_box sides|bounding box (custom count)"]
    bb_type=UI.inputbox prompts, defaults, list, "Bounding Box Control Points Placing"
    case bb_type.to_s
      when "bounding box corners"
        corner_no=0
        while corner_no<=7
          corner=@geom_group.bounds.corner(corner_no)
          control_c_pnt=@entities.add_cpoint corner
          control_c_pnt.set_attribute "on_change_event_params","@geom_groupID",@geom_group.entityID
          control_c_pnt.set_attribute "on_change_event_params","@pwr",@pwr
          control_c_pnt.set_attribute "on_change_event_params","@leave_init_grp",@leave_init_grp
          control_c_pnt.add_observer(Lss_C_Pnts_Group_EntityObserver.new(control_c_pnt))
          corner_no+=1
        end
        self.set_attrs
      when "centers of b_box sides"
        corners_arr=Array.new
        corner_no=0
        while corner_no<=7
          corner=@geom_group.bounds.corner(corner_no)
          corners_arr+=[corner]
          corner_no+=1
        end
        centers_arr=Array.new
        centers_arr+=[Geom::Point3d.linear_combination(0.5, corners_arr[0], 0.5, corners_arr[3])]
        centers_arr+=[Geom::Point3d.linear_combination(0.5, corners_arr[0], 0.5, corners_arr[5])]
        centers_arr+=[Geom::Point3d.linear_combination(0.5, corners_arr[0], 0.5, corners_arr[6])]
        centers_arr+=[Geom::Point3d.linear_combination(0.5, corners_arr[7], 0.5, corners_arr[4])]
        centers_arr+=[Geom::Point3d.linear_combination(0.5, corners_arr[7], 0.5, corners_arr[2])]
        centers_arr+=[Geom::Point3d.linear_combination(0.5, corners_arr[7], 0.5, corners_arr[1])]
        centers_arr.each{|cntr|
          control_c_pnt=@entities.add_cpoint cntr
          control_c_pnt.set_attribute "on_change_event_params","@geom_groupID",@geom_group.entityID
          control_c_pnt.set_attribute "on_change_event_params","@pwr",@pwr
          control_c_pnt.set_attribute "on_change_event_params","@leave_init_grp",@leave_init_grp
          control_c_pnt.add_observer(Lss_C_Pnts_Group_EntityObserver.new(control_c_pnt))
        }
        self.set_attrs
      when "bounding box (custom count)"
        prompts=["Width steps count: ", "Height steps count: ", "Depth steps count: "]
        defaults=[2,2,2]
        wdt,hgt,dpt=UI.inputbox prompts, defaults, "Steps Count"
        corners_arr=Array.new
        corner_no=0
        while corner_no<=7
          corner=@geom_group.bounds.corner(corner_no)
          corners_arr+=[corner]
          corner_no+=1
        end
        pts_arr=Array.new
        wdt_crd=0.0
        while wdt_crd<=wdt
          hgt_crd=0.0
          while hgt_crd<=hgt
            pt1=corners_arr[0]
            pt2=corners_arr[1]
            pt3=corners_arr[2]
            coeffx=wdt_crd/wdt
            coeffy=hgt_crd/hgt
            ptx=Geom::Point3d.linear_combination(1-coeffx, pt1, coeffx, pt2)
            pty=Geom::Point3d.linear_combination(1-coeffy, pt1, coeffy, pt3)
            pt=Geom::Point3d.new(ptx.x,pty.y,corners_arr[0].z)
            pts_arr+=[pt]
            pt=Geom::Point3d.new(ptx.x,pty.y,corners_arr[4].z)
            pts_arr+=[pt]
            hgt_crd+=1.0
          end
          wdt_crd+=1.0
        end
        
        wdt_crd=0.0
        while wdt_crd<=wdt
          dpt_crd=0.0
          while dpt_crd<=dpt
            pt1=corners_arr[0]
            pt2=corners_arr[1]
            pt3=corners_arr[4]
            coeffx=wdt_crd/wdt
            coeffy=dpt_crd/dpt
            ptx=Geom::Point3d.linear_combination(1-coeffx, pt1, coeffx, pt2)
            pty=Geom::Point3d.linear_combination(1-coeffy, pt1, coeffy, pt3)
            pt=Geom::Point3d.new(ptx.x,corners_arr[0].y,pty.z)
            pts_arr+=[pt]
            pt=Geom::Point3d.new(ptx.x,corners_arr[7].y,pty.z)
            pts_arr+=[pt]
            dpt_crd+=1.0
          end
          wdt_crd+=1.0
        end
        
        hgt_crd=0.0
        while hgt_crd<=hgt
          dpt_crd=0.0
          while dpt_crd<=dpt
            pt1=corners_arr[0]
            pt2=corners_arr[2]
            pt3=corners_arr[4]
            coeffx=hgt_crd/hgt
            coeffy=dpt_crd/dpt
            ptx=Geom::Point3d.linear_combination(1-coeffx, pt1, coeffx, pt2)
            pty=Geom::Point3d.linear_combination(1-coeffy, pt1, coeffy, pt3)
            pt=Geom::Point3d.new(corners_arr[0].x,ptx.y,pty.z)
            pts_arr+=[pt]
            pt=Geom::Point3d.new(corners_arr[1].x,ptx.y,pty.z)
            pts_arr+=[pt]
            dpt_crd+=1.0
          end
          hgt_crd+=1.0
        end
        pts_arr.collect! {|pt| pt.to_a}
        pts_arr.uniq!
        
        pts_arr.each{|pt|
          control_c_pnt=@entities.add_cpoint pt
          control_c_pnt.set_attribute "on_change_event_params","@geom_groupID",@geom_group.entityID
          control_c_pnt.set_attribute "on_change_event_params","@pwr",@pwr
          control_c_pnt.set_attribute "on_change_event_params","@leave_init_grp",@leave_init_grp
          control_c_pnt.add_observer(Lss_C_Pnts_Group_EntityObserver.new(control_c_pnt))
        }
        self.set_attrs
    end
    Sketchup.active_model.select_tool nil
  end
  
end #class Lss_Add_C_Pnts_BB

selection=Sketchup.active_model.selection
Sketchup.add_observer(Lss_C_Pnts_AppObserver.new)
if( not file_loaded?("ctrlpnts.rb") )
    tool_menu = UI.menu "Plugins" 
    lss_c_pt_interactive_tool = Lss_C_Pnt_Interactive.new 
    lss_c_pt_interactive_tool.c_pt_interactive_tool=lss_c_pt_interactive_tool
    tool_menu.add_item("Interact with control points...") {
      Sketchup.active_model.select_tool lss_c_pt_interactive_tool
    }

    UI.add_context_menu_handler do |context_menu| 
      if selection.count>1
       context_menu.add_separator 
       context_menu.add_item("Add control points to the curve") {
          chk_curve=selection[0].curve
          if chk_curve
            prompts = ["Add control points to ", "Curve vertices 'behaviour': "]
            defaults= ["each vertex","linear"]
            list = ["each vertex|each 2-nd vertex|each 3-rd vertex|each n-th vertex|start & end vertices", "linear|cosine smoothed"]
            put_c_pnts,pnt_follow_type=UI.inputbox prompts, defaults, list, "Curve Control Points Settings"
            case put_c_pnts
              when "each vertex"
                skip_vert_cnt=0
              when "each 2-nd vertex"
                skip_vert_cnt=1
              when "each 3-rd vertex"
                skip_vert_cnt=2
              when "each n-th vertex"
                skip_vert_cnt=(UI.inputbox ["'n' value: "], [4], "Each N-th Vertex Definition")[0]-1
              when "start & end vertices"
                skip_vert_cnt=-1
            end
            lss_ctrl_pnts=Lss_Ctrl_Pnts.new(skip_vert_cnt,pnt_follow_type,chk_curve)
            lss_ctrl_pnts.add_c_pnts
          else
            UI.messagebox "Selected edges are not parts of single curve."
            return
          end
        }
      end
    end
    
    UI.add_context_menu_handler do |context_menu| 
      if selection.count==1 and selection[0].typename=="Group"
       context_menu.add_separator 
       context_menu.add_item("Add control points to group manually...") {
          Sketchup.active_model.select_tool Lss_Ctrl_Pnts_Add_Tool.new
        }
        context_menu.add_item("Add control points to group bounds...") {
          Sketchup.active_model.select_tool Lss_Add_C_Pnts_BB.new
        }
      end
    end
    
    UI.add_context_menu_handler do |context_menu| 
      if selection.count>0
        allow_add_menu_item=false
       selection.each{|ent| 
          if ent.typename=="ConstructionPoint"
            allow_add_menu_item=true
          end
        }
        if allow_add_menu_item
         context_menu.add_separator 
         context_menu.add_item("Select the group to be under control...") {
            Sketchup.active_model.select_tool Lss_Ctrl_Pnts_Sel_Tool.new
            Sketchup.set_status_text("Click on group you wish to be controlled by selected c-points")
          }
        end
      end
    end
  end
  file_loaded("ctrlpnts.rb")
